#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import time
from .common import *
from .osconnect import getServers, getServerInfo, Token
from datetime import datetime
from .storage import Storage, remove_unneeded_data
from .runcommand import runcommand_e
import argparse
from .configuration import Configuration
import os 
from .version import VERSION

# The main loop interval (in seconds, 1 second is fine)
MAIN_LOOP_INTERVAL = 1

# The idea is to be kind with the system: make monitorings of MONITORING_VM_BLOCK_TIME seconds, monitoring up to MONITORING_VM_BLOCKSIZE VMs
#   and then wait MONITORING_VM_INTERVAL before trying to monitor the next VM block

class Monitor:
    def __init__(self, db):
        self._token = Token()
        self._pending_vms = []
        self._db = db

    def check(self):
        return self._token.renewIfNeeded()

    def updateVMs(self):
        servers = getServers(self._token)
        if servers is None:
            p_error("failed to obtain servers from OpenStack")
            return False
            
        p_debug("{} servers obtained".format(len(servers)))

        for x in servers:
            id = x["id"]
            if id not in self._pending_vms:
                self._pending_vms.append(id)

    def havePendingVMs(self):
        return len(self._pending_vms)

    def monitorVMs(self, limit = None, walltime = 0, conflicting_alert = False, filter_fnc = lambda x: x):
        count = 0
        failed = 0

        p_debugv("{} VMs pending of monitorisation".format(len(self._pending_vms)))

        t0 = datetime.now().timestamp()

        while (count < limit) and (len(self._pending_vms) > 0) and (walltime > 0):
            nextId = self._pending_vms.pop(0)

            # Get the VM info
            serverInfo = getServerInfo(self._token, nextId)
            if serverInfo is None:
                failed += 1
                p_error("failed to obtain VM info from OpenStack")

            # Get the VM status
            if conflicting_alert and ('conflictingRequest' in serverInfo):
                p_warning("{} has conflicting request".format(nextId))
                serverInfo = None
            
            if serverInfo is not None:
                self._db.savevm(nextId, filter_fnc(serverInfo))

            count += 1
            t1 = datetime.now().timestamp()
            walltime = walltime - (t1 - t0)
            p_debugv("still have {} seconds to monitor in this block".format(walltime))
            t0 = t1

        p_debug("{} VMs monitored".format(count))
        return failed

def osidle_monitor():
    config = Configuration({
            "DEFAULT": {
                # The amount of VMs to retrieve at once (default: 25)
                "MONITORING_VM_BLOCKSIZE": 25,
                # The amount of time that can be dedicated to monitor a block of VMs (default: 5)
                "MONITORING_VM_BLOCK_TIME": 5,
                # The amount of time that the monitor waits before the next monitorization (default: 5)
                "MONITORING_VM_INTERVAL": 5,
                # The amount of time that the monitor waits before updating the list of running VMs (default: 600)
                "MONITORING_FULL_INTERVAL": 600,
                # The main loop interval (in seconds, 1 second is fine)
                "MAIN_LOOP_INTERVAL": 1,
                # The amount of time that the monitor waits before notifying for errors, again (default: 21600; i.e. 6 hours)
                "NOTIFICATION_INTERVAL": 21600,
                # The command that will be used to notify the admin for errors in the monitor; the errors will be passed in the standard input (default: None)
                "NOTIFICATION_COMMAND": ("", lambda x: None if x == "" else x),
                # A message to show that the monitor is running (default: "osidled is running")
                "STILL_ALIVE_MESSAGE": "--- osidled is running",
                # The amount of time that the monitor waits before showing a message that the monitor is still running (default: 14400; i.e. 4 hours)
                "STILL_ALIVE_MESSAGE_INTERVAL": 14400,
                # The database file where the monitor will store the data (default: /var/lib/osidled/osidled.db)
                "DATABASE": "/var/lib/osidled/osidled.db",
                # Run the monitor in verbose mode, showing more messages (default: False)
                "VERBOSE": False,
                # Run the monitor in debug mode, showing much more messages (default: False)
                "DEBUG": False,
                # The username to connect to the OpenStack server (default: None)
                "OS_USERNAME": ("", lambda x: None if x == "" else x),
                # The password to connect to the OpenStack server (default: None)
                "OS_PASSWORD": ("", lambda x: None if x == "" else x),
                # The authentication URL to connect to the OpenStack server (default: None)
                "OS_AUTH_URL": ("", lambda x: None if x == "" else x),
                # In case that a VM is not running, OpenStack returns a "conflicting error" and osidle monitor will notify about them. This option allows to ignore such errors (default: False)
                "SILENCE_CONFLICTING": False,
                # By default, osidle monitor will store the raw data obtained for each VM. Changing this option allows to store only the data that will use osidle (default: True, to store the raw data)
                "STORE_RAW_DATA": True,
            }
        }
    )

    parser = argparse.ArgumentParser(allow_abbrev=False, add_help = False)
    parser.add_argument("-c", "--config", default=None, dest="configurationfile", help = "Use this configuration file")

    configargs, _ = parser.parse_known_args()

    if configargs.configurationfile is not None:
        (configfiles, configuration) = config.read([ configargs.configurationfile ])
        if len(configfiles) == 0:
            p_error("failed to read configuration file")
            exit(1)
    else:
        (configfiles, configuration) = config.read([ "./osidled.conf", "/etc/osidle/osidled.conf", "/etc/osidled.conf", "/etc/default/osidled.conf" ], False)

    configuration = configuration["DEFAULT"]
    parser.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS, help='Show this help message and exit.')
    parser.add_argument("-b", "--block-size", dest="blocksize", default=configuration["MONITORING_VM_BLOCKSIZE"], help="The maximum number of VMs to be monitored in a row", type=int)
    parser.add_argument("-t", "--block-time", dest="blocktime", type=int, default=configuration["MONITORING_VM_BLOCK_TIME"], help="The maximum time to be used to monitor a block of VMs")
    parser.add_argument("-C", "--cooldown", dest="cooldown", default=configuration["MONITORING_VM_INTERVAL"], type=int, help="Once a monitoring interval has finished, the amount of seconds to wait before the next monitorisation")
    parser.add_argument("-u", "--vm-update", dest="vmupdate", default=configuration["MONITORING_FULL_INTERVAL"], type=int, help="Time between full updates of the VM list and starting the monitorization of the VMs again")
    parser.add_argument("-l", "--main-loop", dest="mainloop", default=configuration["MAIN_LOOP_INTERVAL"], type=int, help="Main loop interval; the default is {} second and it is not advisable to change it".format(MAIN_LOOP_INTERVAL))
    parser.add_argument("-N", "--notification", dest="notification", type=str, default=configuration["NOTIFICATION_COMMAND"], help="Command to be executed when an error happens; the error message will be passed in the stdin")
    parser.add_argument("-d", "--database", dest="database", help="database to use (at this point, only is valid a sqlite3 file)", default=configuration["DATABASE"])
    parser.add_argument("-s", "--silence-conflicting", dest="conflicting", help="Hide the 'conflicting request' alerts", action="store_false", default=configuration["SILENCE_CONFLICTING"])
    parser.add_argument("-v", "--verbose", dest="verbose", help="verbose mode", action="store_true", default=configuration["VERBOSE"])
    parser.add_argument("-vv", "--verbose-more", dest="verbosemore", help="verbose more mode", action="store_true", default=configuration["DEBUG"])
    parser.add_argument("-U", "--os-username", dest="username", help="OpenStack username (if not set, will be obtained using OS_USERNAME env var)", default=configuration["OS_USERNAME"])
    parser.add_argument("-P", "--os-password", dest="password", help="OpenStack password  (if not set, will be obtained using OS_PASSWORD env var)", default=configuration["OS_PASSWORD"])
    parser.add_argument("-H", "--os-auth", dest="keystone", help="OpenStack keytsone authentication endpoint (if not set, will be obtained using OS_AUTH_URL env var)", default=configuration["OS_AUTH_URL"])
    parser.add_argument("-r", "--remove-unneeded-data", dest="storerawdata", help="Remove the unneeded data before storing it", action="store_false", default=configuration["STORE_RAW_DATA"])
    parser.add_argument('--version', action='version', version=VERSION)

    try:
        args = parser.parse_args()
    except argparse.ArgumentError as e:
        sys.exit(1)

    if args.verbose:
        setVerbose(1)
    if args.verbosemore:
        setVerbose(2)

    if len(configfiles) > 0:
        p_debug("Reading configuration from {}".format(", ".join(configfiles)))

    if args.username is not None:
        os.environ['OS_USERNAME'] = args.username
    if args.password is not None:
        os.environ['OS_PASSWORD'] = args.password
    if args.keystone is not None:
        os.environ['OS_AUTH_URL'] = args.keystone

    # Connect to the database (if possible)
    p_debugv("connecting to the database {}".format(args.database))
    storage = Storage(args.database)
    storage.connect()

    # Prepare the monitor
    monitor = Monitor(storage)

    if not monitor.check():
        p_error("failed to connect to OpenStack")
        sys.exit(1)

    # Start the main loop
    t0_vm = 0
    t0_full = 0
    t00 = datetime.now().timestamp()
    p_info("starting the monitoring loop")
    tN_0 = 0
    filter_fnc = lambda x: x

    # List of errors
    error_list = []

    if not args.storerawdata:
        p_debug("removing the unneeded data")
        filter_fnc = remove_unneeded_data

    while True:
        try:
            t1 = datetime.now()
            t1 = t1.timestamp()

            if len(error_list) > 0:
                if t1 - tN_0 > configuration["NOTIFICATION_INTERVAL"]:
                    tN_0 = t1

                    if args.notification is not None:
                        p_debug("notifying using command {}".format(args.notification))
                        retcode, cout, cerr = runcommand_e(args.notification, strin = "\n".join(error_list).encode("utf-8"))
                        p_debug("notification command returned {}".format(retcode))
                        p_debugv("notification command stdout: {}".format(cout))
                        p_debugv("notification command stderr: {}".format(cerr))

                    p_error("{} errors happened during the last {} seconds".format(len(error_list), configuration["NOTIFICATION_INTERVAL"]))
                    error_list = []

            if (t0_vm + args.blocktime) < t1:
                if monitor.havePendingVMs():
                    p_debug("monitorising up to {} vms".format(args.blocksize))

                    failed = monitor.monitorVMs(args.blocksize, args.blocktime, args.conflicting, filter_fnc)
                    if failed > 0:
                        p_debug("failed to monitor {} vms".format(failed))
                        error_list.append(s_error("failed to monitor {} vms".format(failed)))

                t0_vm = datetime.now().timestamp()

            if (t0_full + args.vmupdate) < t1:
                p_debug("full monitorisation")

                monitor.updateVMs()

                t0_vm = datetime.now().timestamp()
                t0_full = t0_vm

            if t1 - t00 > configuration["STILL_ALIVE_MESSAGE_INTERVAL"]:
                p_info(configuration["STILL_ALIVE_MESSAGE"])
                t00 = t1

        except Exception as e:
            p_error('Error: {}'.format(e))
            error_list.append(s_error(str(e)))

        p_debugv("wait {}".format(args.mainloop))
        time.sleep(args.mainloop)

if __name__ == "__main__":
    osidle_monitor()


#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import os
import requests
import json
from dateutil.parser import parse as datetimeparser
from .common import *

class Token:
    def __str__(self):
        return "Token: {}\nURL: {}\nExpiration: {}".format(self._token, self._url, self._expiration)

    def __init__(self):
        self._token = None
        self._url = None
        self._expiration = None

    """
    Method to force getting a token from OpenStack
    """
    def get(self):
        passwd = os.environ.get('OS_PASSWORD')
        url_base = os.environ.get('OS_AUTH_URL')
        user = os.environ.get('OS_USERNAME')

        headers = {'Content-Type': 'application/json', "X-OpenStack-Nova-API-Version": "2.48"}
        obj = { "auth": { "scope":"system", "identity": { "methods": [ "password" ], "password": { "user": { "name": user, "domain": { "id": "default" }, "password": passwd } } } } }

        now = datetime.now().timestamp()

        # Make a request to create the token
        try:
            p_debugv("creating a new token at {}/auth/tokens".format(url_base))
            token_req = requests.post("{}/auth/tokens".format(url_base), data=json.dumps(obj), headers=headers, timeout = 5)
        except Exception as e:
            p_error("Could not connect to OpenStack: {}".format(e))
            return False

        # Check if the request was successful
        if (token_req.status_code < 200 or token_req.status_code >= 300):
            p_error("Failed to connect to OpenStack")
            return False

        if 'X-Subject-Token' not in token_req.headers:
            p_error("Failed to obtain the token")
            return False

        token = token_req.headers['X-Subject-Token']

        # Now deal with the content of the request
        token_req = token_req.json()

        # Get the nova endpoint
        nova_ep = None
        for endpoint in token_req["token"]["catalog"]:
            if endpoint["name"] == "nova":
                nova_ep = endpoint
                break

        if nova_ep is None:
            p_error("could not get nova endpoint")
            return False

        admin_ep = None
        for ep in nova_ep["endpoints"]:
            if ep["interface"] == "admin":
                admin_ep = ep
                break

        if admin_ep is None:
            p_error("could not get nova admin endpoint")
            return False

        token_issued = datetimeparser(token_req['token']['issued_at'])
        token_expires = datetimeparser(token_req['token']['expires_at'])

        # Now store the information about the token
        p_debug("Token obtained: {}".format(token))
        self._expiration = datetime.fromtimestamp(now + token_expires.timestamp() - token_issued.timestamp() )
        self._url = admin_ep['url']
        p_debugv("Token expiration at: {}".format(self._expiration))
        p_debugv("Token URL: {}".format(self._url))
        self._token = token
        return True

    def isValid(self):
        if self._token is None:
            return False
        if self._expiration < datetime.now():
            return False
        return True

    def renewIfNeeded(self):
        if not self.isValid():
            return self.get()
        return True

    @property
    def url(self):
        return self._url
    @property
    def token(self):
        return self._token

def tokenQuery(token, query, headers = {}, timeout = 5):
    # Performs a query using the provided token
    # @param token: the token to use
    # @param query: the query to perform
    if not token.renewIfNeeded():
        p_error("could not get a valid token")
        return None
    try:
        query = query.lstrip("/")
        req = requests.get(f"{token.url}/{query}", headers = {**headers, 'X-Auth-Token': token.token }, timeout = timeout)
    except Exception as e:
        print(e)
        p_error("Could not connect to OpenStack")
        return None

    return req.json()

def autoQuery(query, headers = {}):
    # Performs a query using a token obtained from the environment variables
    # @param query: the query to perform
    token = Token()
    if not token.get():
        return None
    return tokenQuery(token, query, headers)

def getServers(token):
    # Gets the information regarding to the servers, using the provided token
    # @param token: the token to use
    # @return: a list of dictionaries, each one containing the information about a server or None if an error happened
    servers = tokenQuery(token, "/servers?all_tenants=1")
    if servers is None:
        return None
    return servers["servers"]

def getServerInfo(token, id):
    # Gets the information regarding to the server with the given id, using the provided token
    # @param token: the token to use
    # @param id: the id of the server
    # @return: a json object containing the information about the server or None if an error happened
    return tokenQuery(token, f"/servers/{id}/diagnostics", headers = {"X-OpenStack-Nova-API-Version": "2.48"}, timeout = 15)

def checkConnection(token):
    # Checks if the token is still valid
    # @param token: the token to use
    # @return: True if the token is still valid, False otherwise
    if not token.renewIfNeeded():
        p_error("could not get a valid token")
        return False

    return True



# coding: utf-8
#
# CLUES Python utils - Utils and General classes that spin off from CLUES
# Copyright (C) 2017 - GRyCAP - Universitat Politecnica de Valencia
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

import subprocess
import logging
import time
import threading
import os
import signal

class CommandError(Exception):pass

# TODO: in case that we try to implement the simulator, we should reimplement the run_command apps to use timemachine and events
#       * it probably will not be needed, because the commands are supposed to not to fail in case that we are simmulating; this
#         this polling method is for production purposes
def _runcommand(command, shell=False, timeout = None, strin = None, cwd = None):
    try:
        p = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, cwd = cwd, stderr=subprocess.PIPE, shell=shell, preexec_fn = os.setsid)
    except Exception as e:
        if type(command)==list: command = " ".join(command)
        logging.error('Could not execute command "%s"' %command)
        raise

    timer = threading.Timer(timeout, os.killpg, [p.pid, signal.SIGKILL])
    timer.start()
    (out, err) = p.communicate(input=strin)
    timer.cancel()
    return (p.returncode, out, err)

def runcommand(command, shell = True, timeout = None, strin = None, cwd = None):
    cout = ""
    try:
        retcode, cout, cerr = _runcommand(command, shell, timeout, strin, cwd)
        if retcode != 0:
            if type(command)==list: command = " ".join(command)
            logging.error(' Error in command "%s"' % command)
            logging.error(' Return code was: %s' % retcode)
            logging.error(' Error output was:\n%s' % cerr)
            return False, cout
        else:
            return True, cout
    except OSError:
        logging.error("error executing command %s" % command)
        return False, cout
    return True, cout

def runcommand_e(command, shell = True, timeout = None, strin = None, cwd = None):
    cout = ""
    cerr = ""
    try:
        retcode, cout, cerr = _runcommand(command, shell, timeout, strin, cwd)
    except OSError:
        logging.error("error executing command %s" % command)
        return -1, cout, cerr
    return retcode, cout, cerr



#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import copy
from datetime import datetime, timedelta
import json

from osidle.common import p_warning

class RawData:
    def __init__(self, data, args):
        _data = RawData._convert(data)
        # TODO: discard the first sample as it is just the historic data?
        _data = _data[1:]
        self._data = RawData._get(_data, args.fromdate, args.todate)
        self._ncpu = None
        self._nnic = None
        self._ndisk = None

        if len(self._data) > 0:
            self._ncpu = self._data[0]["ncpu"]
            self._nnic = self._data[0]["nnic"]
            self._ndisk = self._data[0]["ndisk"]

    @property
    def ncpu(self):
        return self._ncpu

    @property
    def nnic(self):
        return self._nnic

    @property
    def ndisk(self):
        return self._ndisk
        
    def __str__(self):
        return self.dumpdata("json", True)

    def dumpdata(self, format, pretty = False, transform_fnc = None):
        if format == "csv" or format == "excel":
            lines = []
            for d in self._data:
                v = (d["T"], d["S"], d["e"], 
                    d["tcpu"], d["tdisk"], d["tnic"],
                    *d["cpu"], *[ v for x in d["disk"] for v in [ x["r"], x["w"] ] ] , # The CPUs
                    *[ v for x in d["nic"] for v in [ x["rx"], x["tx"] ] ]) # The NICs  
                if callable(transform_fnc):
                    v = transform_fnc(v)
                lines.append(v)

            if format == "csv":
                separador=","
                return "\n".join([ separador.join([str(v) for v in x ]) for x in lines ])
            else:
                separador = ";"
                return "\n".join([ separador.join([str(v).replace(".", ",") for v in x ]) for x in lines ])
        else:
            if pretty:
                return json.dumps(self._data, indent=4)
            else:
                return json.dumps(self._data)

    @property
    def data(self):
        return self._data

    @staticmethod
    def _get(data, fromDate = None, toDate = None):
        t_fromDate = fromDate.timestamp()
        t_toDate = toDate.timestamp()
        if (t_fromDate > t_toDate):
            return []

        _data = []
        for d in data:
            time_after_begin = max(0, t_fromDate - d["S"])
            time_before_end = max(0, d["S"] + d["e"] - t_toDate)
            e = d["e"] - time_after_begin - time_before_end

            if e == d["e"]:
                _data.append(d)
            else:
                # Duplicate d and update the data
                d = copy.deepcopy(d)
                fraction = e / d["e"]
                d["cpu"] = [ x * fraction for x in d["cpu"] ]
                d["disk"] = [ { "r": x["r"] * fraction, "w": x["w"] * fraction } for x in d["disk"] ]
                d["nic"] = [ { "rx": x["rx"] * fraction, "tx": x["tx"] * fraction } for x in d["nic"] ]
                d["S"] = d["S"] + time_after_begin
                d["e"] = e
                d["T"] = datetime.fromtimestamp(d["S"]).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                d["tcpu"] = d["tcpu"] * fraction
                d["tdisk"] = d["tdisk"] * fraction
                d["tnic"] = d["tnic"] * fraction
                _data.append(d)

        return _data

    # This function gets a data series and converts it to an incremental data series. It assumes that each of the samples in the input
    # data series is an absolute sample and so it converts it to an incremental data series by subtracting the previous sample from the
    # current one.
    # Returns a set of samples with the following data:
    #   - t: the timestamp where the sample was taken; 
    #   - s: t in seconds
    #   - T: the tiemstamp where the sample starts; 
    #   - S: T in seconds
    #   - e: the usage of the resources correspond to this continuous number of seconds
    #   - cpu, nic, disk: the usage of the CPU, NIC and DISK
    #   - ncpu, nnic, ndis: number of CPUs, NICs and DISKs
    #   - tcpu, tnic, tdisk: the aggregated amount of nanoseconds of CPU used, NIC tx and rx and DISK r and w
    @staticmethod
    def _convert(data):
        converted = []
        d0 = None

        for d in data:
            # Skip the samples that do not have information
            if "conflictingRequest" in d:
                continue
            if "itemNotFound" in d:
                continue
            if (not 'cpu_details' in d) or (not 'disk_details' in d) or (not 'nic_details' in d):
                p_warning("Missing information in the sample: {}".format(json.dumps(d)))
                continue

            d["tcpu"] = sum(x["time"] for x in d["cpu_details"]) * 1e-9
            d["tdisk"] = sum([ x["read_bytes"] + x["write_bytes"] for x in d["disk_details"] ])
            d["tnic"] = sum([ x["rx_octets"] + x["tx_octets"] for x in d["nic_details"] ])

            # If it is the first sample, we'll use it as the base but we need to convert it to a "pseudo-incremental" sample that
            # starts the series with values set to 0. It is needed to adjust the timestamp
            if d0 is None:
                # Clone d to create the fake first sample
                d0 = copy.deepcopy(d)

                # t is the timestamp of the sample (when it was obtained)
                t = datetime.strptime(d["t"], "%Y-%m-%dT%H:%M:%S.%fZ") - timedelta(seconds=d["uptime"])
                d0["t"] = t.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
                # T is the timestamp where the sample starts (when the consumption of resources started)
                d0["T"] = d["t"]
                # e is the difference between the previous timestamp and this one
                d0["e"] = 0,
                # s is the timestamp (t) in seconds
                d0["s"] = t.timestamp() # d0["s"] - d0["uptime"]
                # S is the timestamp (s) in seconds
                d0["S"] = d0["s"]
                # Now clear the resource consumption as this is a "base sample" (e.g.) the VM was just started
                d0["uptime"] = 0
                for c in d0["cpu_details"]:
                    c['time'] = 0
                for x in d0["disk_details"]:
                    x['read_bytes'] = 0
                    x['write_bytes'] = 0
                for n in d0["nic_details"]:
                    n['rx_octets'] = 0
                    n['tx_octets'] = 0
                d0["tcpu"] = 0
                d0["tdisk"] = 0
                d0["tnic"] = 0

            # Calculate the current incremental sample by subtracting the previous sample from the current one (also conver cpu from nanoseconds to seconds)
            d1 = {
                "t": d['t'],
                "s": d['s'],
                "ncpu": len(d['cpu_details']),
                "ndisk": len(d['disk_details']),
                "nnic": len(d['nic_details']),
                "tcpu": sum(x["time"] for x in d["cpu_details"]) * 1e-9,
                "tdisk": sum([ x["read_bytes"] + x["write_bytes"] for x in d["disk_details"] ]),
                "tnic": sum([ x["rx_octets"] + x["tx_octets"] for x in d["nic_details"] ]),
            }
            # TODO: what to do with resized instances (either the number of CPUs or the number of NICs or the number of DISKs)

            if d0["tcpu"] > d1["tcpu"] or d0["tdisk"] > d1["tdisk"] or d0["tnic"] > d1["tnic"]:
                # The VM has been stopped and started again later; will not subtract the previous sample from the current one

                # We'll skip the sample, because we do not know the time from the previous sample: the requests between this and the previous one
                #   are conflicting and we cannot calculate the incremental consumption.

                # TODO: When a VM is stopped and started later, the amount of CPU is reset to 0...
                #   the same happens to disk and nic, but not the uptime. 
                p_warning("Skipping sample with negative number of CPUs, disks or NICs (the VM was probably shut down)")
            else:
                d1["tcpu"] = d1["tcpu"] - d0["tcpu"]
                d1["tdisk"] = d1["tdisk"] - d0["tdisk"]
                d1["tnic"] = d1["tnic"] - d0["tnic"]

                # TODO: now we have to calculate the values of T, S and e both for d and for d1
                # So we have to calculate the elapsed time, which is the difference between the timestamp of the previous sample and the current one
                d1["e"] = d["s"] - d0["s"]
                d1["S"] = d["s"] - d1["e"]
                d1["T"] = datetime.fromtimestamp(d1["S"]).strftime("%Y-%m-%dT%H:%M:%S.%fZ")

                # d1["tcpu"] = sum(d1["cpu"])
                # d1["tdisk"] = sum([ x for n in d1["disk"] for x in [n["r"], n["w"]]])
                # d1["tnic"] = sum([ x for n in d1["nic"] for x in [n["rx"], n["tx"]]])

                # Add the incremental sample to the result
                converted.append(d1)

            # Store the sample to be the reference for the previous one
            d0 = d

        return converted



#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from distutils.command.config import config
from setuptools import setup
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

from osidle.version import VERSION

# The data files to be installed. The syntax is the same as for the setup(), but in this version
#   it is possible to chown the files and to force overwritting or not:
#   ( "destination folder", [ "list of relative filenames or folders to be copied to the destination folder", "user:group" to chown, True/False to force overwritting ] )
#   - The user:group can be set to "" to not to chown the files, or omit any of them to not to chown the group or the user
data_files=[
  # Copy the base configuration file to the global folder
  ('/etc/osidled/', ['etc/osidled.conf'], "", False),
  ('/etc/osidled/', ['etc/osidle-notify'], "", True),
  ('/etc/default', ['etc/osidled.conf'], "", True),
  # Prepare the service configuration file
  ('/etc/systemd/system', ['etc/systemd/system/osidled.service'], "", True),
  # Make sure that the working folder for the service is created
  ('/var/lib/osidled/', [], "", False),
]

def chown(path, user, recursive=False):
  if user == "":
    return

  user_p = user.split(":")
  user = user_p[0] if user_p[0] != "" else None
  group = user_p[1] if len(user_p) > 1 else None

  import pwd
  import grp

  try:
    uid = pwd.getpwnam(user).pw_uid
  except KeyError:
    uid = None
  
  try:
    gid = grp.getgrnam(group).gr_gid
  except KeyError:
    gid = None

  try:
    if not recursive or os.path.isfile(path):
      shutil.chown(path, user, group)
    else:
      for root, dirs, files in os.walk(path):
        shutil.chown(root, user, group)
        for item in dirs:
          shutil.chown(os.path.join(root, item), user, group)
        for item in files:
          shutil.chown(os.path.join(root, item), user, group)
  except OSError as e:
    raise e 

if __name__ == "__main__":
  import os
  import shutil
  import sys
  import argparse

  parser = argparse.ArgumentParser(description='osidle setup', allow_abbrev=False)
  parser.add_argument('--overwrite-config', dest="overwrite_config", action='store_true', help='Overwrite the configuration files if exist', default=False)
  args, unkown = parser.parse_known_args()
  overwriteconfig = False
  if args.overwrite_config:
    sys.argv.remove('--overwrite-config')
    overwriteconfig = True

  def copyfiles():
    global data_files, overwriteconfig
    for options in data_files:
      if len(options) < 2:
        raise Exception("Invalid data file options")
      if len(options) < 3:
        options = [ *options, "" ]
      if len(options) < 4:
        options = [ *options, False ]
      [ folder, files, user, forceoverwrite ] = options
      # Get the user to change the ownership of the files
      # Create the folder if it does not exist
      if not os.path.exists(folder):
        os.makedirs(folder)
        chown(folder, user, recursive=False)
      if not os.path.isdir(folder):
        raise Exception("Error: {} is not a directory\n".format(folder))
      for file in files:
        configfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), file)
        if os.path.isdir(configfile):
          # It is a folder... so copy recursively
          dest_folder = os.path.join(folder, file)
          if forceoverwrite or overwriteconfig or not os.path.exists(dest_folder):
            shutil.copytree(configfile, folder)
            chown(dest_folder, user, recursive=True)
        elif os.path.isfile(configfile):
          # It is a file... so copy it
          dest_filename = os.path.join(folder, os.path.basename(file))
          if forceoverwrite or overwriteconfig or (not os.path.exists(dest_filename)):
            shutil.copy2(configfile, folder)
            chown(dest_filename, user, recursive=False)
        else:
          raise Exception("Error: could not find file {}\n".format(file))            

  class PostInstallCommand(install):
    def run(self):
      super().run()
      copyfiles()

  class PostDevelopCommand(develop):
    def run(self):
      super().run()
      copyfiles()

  setup(
    name = 'osidle',            
    packages = ['osidle'],
    version = VERSION,          
    license='Apache 2.0',                   # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description = 'A system that gathers information for the usage of resources of the VMs in OpenStack, to try to detect which of them are idle',
    long_description = README,
    long_description_content_type = 'text/markdown',
    author = 'Carlos A.',             
    author_email = 'caralla@upv.es',  
    url = 'https://github.com/grycap/osidle',
    keywords = ['openstack', 'idle', 'virtual machine' ],
    install_requires=[
            'dateutils',
            'requests',
            'tqdm',
            'xlsxwriter'
        ],
    cmdclass={
        'install': PostInstallCommand,
        'develop': PostDevelopCommand,
    },
    classifiers=[
      'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
      'Intended Audience :: System Administrators',
      'Topic :: Utilities',
      'License :: OSI Approved :: Apache Software License',
      'Programming Language :: Python :: 3',
    ],
    entry_points = {
      'console_scripts' : [
        'osidle=osidle.analysis:osidle_analysis',
        'osidled=osidle.monitor:osidle_monitor',
        'osidle-packdb=osidle.packdb:osidle_packdb',
      ]
    },
    # data_files=data_files
  )




#
#    Copyright 2022 - Carlos A. <https://github.com/dealfonso>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import sqlite3
import json
from .common import p_error, p_warning, p_debugv, p_debug, p_info
from datetime import datetime, timedelta

DEFAULT_FILENAME = "monitoring.sqlite3"

def _sql(cursor, query, params = ()):
    p_debugv("{}: {}".format(query,params))
    result = cursor.execute(query, params)
    return result

class Storage:
    def __init__(self, filename = None):
        if filename is None:
            filename = DEFAULT_FILENAME
        self._filename = filename
        self._conn = None

    def connect(self):
        conn = None
        try:
            conn = sqlite3.connect(self._filename)
        except Exception as e:
            p_error("Could not connect to database: {}".format(e))
            conn = None

        if conn is not None:
            self._conn = conn
            self._createDB()
    
    def _createDB(self):
        cursor = self._conn.cursor()
        cursor.execute("create table if not exists \
            vmmonitor (\
                id integer PRIMARY KEY AUTOINCREMENT, \
                t datetime DEFAULT (STRFTIME('%Y-%m-%dT%H:%M:%fZ', 'NOW')), \
                vmid varchar(36) NOT NULL,\
                data text\
            )")
        self._conn.commit()

    def isConnected(self):
        return self._conn is not None

    def getmaxt(self):
        if not self.isConnected():
            return None

        cursor = self._conn.cursor()
        cursor.execute("select t from vmmonitor order by t desc limit 1")
        try:
            return datetime.strptime(cursor.fetchone()[0], "%Y-%m-%dT%H:%M:%S.%fZ")
        except Exception as e:
            # Just in case the DB is not initialized
            return None

    # Obtains the min available timestamp (i.e. the first time that the system was started)
    def getmint(self):
        if not self.isConnected():
            return None

        cursor = self._conn.cursor()
        cursor.execute("select t from vmmonitor order by t asc limit 1")
        try:
            return datetime.strptime(cursor.fetchone()[0], "%Y-%m-%dT%H:%M:%S.%fZ")
        except Exception as e:
            # Just in case the DB is not initialized
            return None

    def savevm(self, vmid, info):
        if not self.isConnected():
            return False

        cursor = self._conn.cursor()
        cursor.execute("insert into vmmonitor (vmid, data) values (?, ?)", (vmid, json.dumps(info)))
        self._conn.commit()
        return True

    def getvmdata(self, vmid, fromDate = None, toDate = None):
        if not self.isConnected():
            return []

        cursor = self._conn.cursor()

        if (fromDate is None) and (toDate is None):
            _sql(cursor, "select t, data from vmmonitor where vmid = ? order by t asc", (vmid,))
            # cursor.execute("select t, data from vmmonitor where vmid = ? order by t asc", (vmid,))
        elif (fromDate is None):
            fromDate = fromDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "select t, data from vmmonitor where vmid = ? and t >= ? order by t asc", (vmid, fromDate))
        elif (toDate is None):
            toDate = toDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "select t, data from vmmonitor where vmid = ? and t <= ? order by t asc", (vmid, toDate))
        else:
            fromDate = fromDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            toDate = toDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "select t, data from vmmonitor where vmid = ? and t >= ? and t <= ? order by t asc", (vmid, fromDate, toDate))
            
        # TODO: filter the data and return the objects in the right format
        result = []
        for (t, data) in cursor.fetchall():
            data = json.loads(data)
            data["t"] = t
            data["s"] = datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%fZ").timestamp()
            result.append(data)
        
        return result
        # return [ {"t": datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%fZ"), "i": json.loads(data)} for (t, data) in cursor.fetchall() ]

    def getvms(self, fromDate = None, toDate = None):
        if not self.isConnected():
            return []

        cursor = self._conn.cursor()
        # cursor.execute("select vmid from vmmonitor group by vmid")

        if (fromDate is None) and (toDate is None):
            _sql(cursor, "select vmid from vmmonitor group by vmid")
            # cursor.execute("select t, data from vmmonitor where vmid = ? order by t asc", (vmid,))
        elif (fromDate is None):
            fromDate = fromDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "select vmid from vmmonitor group by vmid where t >= ?", (fromDate))
        elif (toDate is None):
            toDate = toDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "select vmid from vmmonitor group by vmid where t <= ?", (toDate))
        else:
            fromDate = fromDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            toDate = toDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "select vmid from vmmonitor group by vmid where t >= ? and t <= ?", (fromDate, toDate))

        return [ x for (x,) in cursor ]

    def delete(self, keepFromDate, keepToDate):
        if not self.isConnected():
            return False

        cursor = self._conn.cursor()

        if (keepFromDate is None) and (keepToDate is None):
            raise Exception("refusing to wipe the whole database")
        elif (keepFromDate is None):
            keepToDate = keepToDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "delete from vmmonitor where t > ?", (keepToDate,))
        elif (keepToDate is None):
            keepFromDate = keepFromDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "delete from vmmonitor where t < ?", (keepFromDate,))
        else:
            keepFromDate = keepFromDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            keepToDate = keepToDate.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            _sql(cursor, "delete from vmmonitor where t < ? or t > ?", (keepFromDate, keepToDate))

        self._conn.commit()
        return cursor.rowcount

    def vaccuum(self):
        if not self.isConnected():
            return False

        cursor = self._conn.cursor()
        _sql(cursor, "vacuum")
        self._conn.commit()
        return True

    def filterdata(self, filter_fnc, pre_fnc = None, post_fnc = None):
        if not self.isConnected():
            return False

        cursor1 = self._conn.cursor()

        _sql(cursor1, "select count(*) from vmmonitor")
        count = cursor1.fetchone()[0]

        if (callable(pre_fnc)):
            pre_fnc(count)

        cursor2 = self._conn.cursor()
        _sql(cursor1, "select id, vmid, t, data from vmmonitor")
        for (id, vmid, t, data) in cursor1:
            data = json.loads(data)
            data = filter_fnc(vmid, t, data)
            if data is not None:
                cursor2.execute("update vmmonitor set data = ? where id = ?", (json.dumps(data), id))
            else:
                cursor2.execute("delete from vmmonitor where id = ?", (id, ))

        self._conn.commit()

        if (callable(post_fnc)):
            post_fnc()

        return True

def remove_unneeded_data(data):
    if ('cpu_details' not in data) or ('disk_details' not in data) or ('nic_details' not in data):
        return None
        
    return {
        "cpu_details": [ { "time": x["time"] } for x in data["cpu_details"] ],
        "disk_details": [ { "read_bytes": x["read_bytes"],  "write_bytes": x["write_bytes"]} for x in data["disk_details"] ],
        "nic_details": [ { "mac_address": x["mac_address"], "rx_octets": x["rx_octets"], "tx_octets": x["tx_octets"] } for x in data["nic_details"] ],
        "num_cpus": data["num_cpus"],
        "num_disks": data["num_disks"],
        "num_nics": data["num_nics"],
        "uptime": data["uptime"],
        "state": data["state"],
    }
